searchNodes=[{"ref":"Erps.html","title":"Erps","type":"module","doc":"Erps is an OTP-compliant remote protocol service The general purpose of Erps is to extend the GenServer primitives over public networks using two-way TLS authentication and encryption without using OTP standard distribution, for cases when: Latency, reliability, or network toplogy properties make using the erlang distribution gossip protocol unfavorable or unreliable You prefer to use erlang distributed clusters strictly contain members of a given type, and have an orthogonal channel for cluster-to-cluster communication. You want to keep groups of Erlang nodes in isolation with a restricted protocol language to achieve security-in-depth and prevent privilege escalation across cluster boundaries. Erps is an asymmetric protocol, which means there are distinct roles for client and server. An Erps client is a &#39;dumb proxy&#39; for a remote GenServer, forwarding all of its call/2 and cast/2 requests to the Erps server. Erps is also a persistent protocol. The clients are intended to connect in to the target server and indefinitely issue requests over this channel, (more like websockets) instead of instantiating a new connection per request, (like HTTP, JSON API, GRPC, or GraphQL). To that end, the Erps Client and Daemon support self-healing connections by default. Testability To make working with your program easier in developer or staging environments, Erps can be configured to use TCP. In order to seamlessly switch between these transport modules, Erps uses the Transport library; both clients and servers may be launched with transport: Transport.Tcp to use this mode. Security model Erps currently presumes that once authenticated via TLS, both ends of the Erps connection are trusted, and provides only basic security measures to protect the server and the client. Distribution and Worker pools There are no provisions for supporting out-of-the-box &#39;smart&#39; distribution defaults or smart worker pool defaults, but this is planned. Multiverses Erps supports Multiverses out of the box. To activate multiverses at compile- time, you will probably want to have the following in your test.exs: config :erps, use_multiverses: true In order to use multiverses with any given Daemon or Client, you must also pass forward_callers: true in the options of the Erps.Daemon.start/3, Erps.Daemon.start_link/3, Erps.Client.start/3, Erps.Client.start_link/3 functions. Examples Examples on how to set up basic clients and servers are provided in in the documentation for Erps.Client and Erps.Server. On the server side you must use an Erps.Daemon to serve as the entry-point for multiple connections which are individually managed by servers; A Server cannot listen on a port for an inbound connection by itself."},{"ref":"Erps.html#is_remote/1","title":"Erps.is_remote/1","type":"macro","doc":"Tests to see if a GenServer.from/0 tuple being passed into an Erps.Server is from a remote client usable in guards."},{"ref":"Erps.Client.html","title":"Erps.Client","type":"behaviour","doc":"Create an Erps client GenServer. The best way to think of an Erps client is that it is a GenServer that forwards its call/2 and cast/2 callbacks to a remote GenServer over a LAN or WAN. This callbacks would normally be provided by standard GenServer.call/2 and GenServer.cast/2 semantics over erlang distribution but sometimes you just don&#39;t want that (see Erps). Basic operation Presuming you have set up an Erps server GenServer on some host at @hostname, you can connect the client and the server simply by instantiating the server module. Example defmodule ErpsClient do use Erps.Client @hostname &lt;...&gt; @port &lt;...&gt; def start_link, do: Erps.Client.start_link(__MODULE__, :ok, server: @hostname, port: @port, tls_opts: [...]) def init(init_state), do: {:ok, init_state} end {:ok, client} = ErpsClient.start_link GenServer.call(client, :some_remote_call) # =&gt; :some_remote_response Module options :version the version of your Erps API messages. Should be a SemVer string. see Version for more information. :identifier (optional) a binary identifier for your Erps API endpoint. Maximum 36 bytes, suggested to be human-readable. This must match the identifier on the server in order for there to be a successful connection. :safe (see :erlang.binary_to_term/2), for decoding terms. If set to false, then allows undefined atoms and lambdas to be passed via the protocol. This should be used with extreme caution, as disabling safe mode can be an attack vector. (defaults to true) Example defmodule MyClient do use Erps.Client, version: &quot;0.2.4&quot;, identifier: &quot;my_api&quot;, safe: false def start_link(iv) do Erps.Client.start_link(__MODULE__, init, server: &quot;my_api-server.example.com&quot;, port: 4747, transport: Transport.Tls, tls_opts: [...]) end def init(iv), do: {:ok, iv} end Important Notes Calls are non-blocking for the Erps clients (and can be for the servers, if you so implement them). You may issue multiple, asynchronous calls across the network from different processes and they will be routed correctly and will only interfere with each other in terms of the connection arbitration overhead. Calls reply immediately with the atom :disconnected if the erps client is not currently connected to its server. handle_continue/2 is currently not supported. This is a limitation in the Connection library."},{"ref":"Erps.Client.html#connect/1","title":"Erps.Client.connect/1","type":"function","doc":"triggers a connection. Only use this function after using disconnect/1"},{"ref":"Erps.Client.html#disconnect/1","title":"Erps.Client.disconnect/1","type":"function","doc":"triggers a disconnection. Useful to silence persistent Erps clients during maintenance phases. Note: this might cause upstream errors as consumers of the Erps service will emit errors on calls and casts."},{"ref":"Erps.Client.html#c:handle_connect/2","title":"Erps.Client.handle_connect/2","type":"callback","doc":"called when a connection has been successfully established. Return codes see return codes for handle_push/2"},{"ref":"Erps.Client.html#c:handle_info/2","title":"Erps.Client.handle_info/2","type":"callback","doc":"Invoked to handle general messages sent to the client process. Most useful if the client needs to be attentive to system messages, such as nodedown or monitored processes, but also useful for internal timeouts. see: GenServer.handle_info/2. Return codes see return codes for handle_push/2"},{"ref":"Erps.Client.html#c:handle_push/2","title":"Erps.Client.handle_push/2","type":"callback","doc":"Invoked to handle Erps.Server.push/2 messages. push is the push message sent by a Erps.Server.push/2 and state is the current state of the Erps.Client. Return codes {:noreply, new_state} continues the loop with new state new_state {:noreply, new_state, timeout} causes a :timeout message to be sent to handle_info/2 if no other message comes by {:noreply, new_state, :hibernate}, causes a hibernation event (see :erlang.hibernate/3) {:disconnect, reason, new_state}, causes a disconnection. {:stop, reason, new_state} terminates the loop, passing new_state to terminate/2, if it&#39;s implemented."},{"ref":"Erps.Client.html#c:init/1","title":"Erps.Client.init/1","type":"callback","doc":"Invoked to set up the process. Like GenServer.init/1, this function is called from inside the process immediately after start_link/3 or start/3. Return codes {:ok, state} a succesful startup of your intialization logic and sets the internal state of your server to state. {:ok, state, timeout} the above, plus a :timeout atom will be sent to handle_info/2 if no other messages come by. {:ok, state, :hibernate} successful startup, followed by a hibernation event (see :erlang.hibernate/3) :ignore - Drop the gen_server creation request, because for some reason it shouldn&#39;t have started. {:stop, reason} - a failure in creating the gen_server. Results in {:error, reason} being propagated as the result of the start_link"},{"ref":"Erps.Client.html#socket/1","title":"Erps.Client.socket/1","type":"function","doc":"returns the current socket in use by the server. This may be a TCP socket or an SSL socket, or another interface depending on what transport strategy you&#39;re using."},{"ref":"Erps.Client.html#start/3","title":"Erps.Client.start/3","type":"function","doc":"starts a client GenServer, not linked to the caller. Most useful for tests. see start_link/3 for a description of avaliable options."},{"ref":"Erps.Client.html#start_link/3","title":"Erps.Client.start_link/3","type":"function","doc":"starts a client GenServer, linked to the caller. Will attempt to contact the server over the specified transport strategy. If the connection fails, the client will be placed in an invalid state until connection succeeds, with a reconnect interval specified in the module options. options :server IP address of the target server (required) :port IP port of the target server (required) :transport module for communication transport strategy :keepalive time interval for sending a TCP/IP keepalive token. :tls_opts options for setting up a TLS connection. :cacertfile path to the certificate of your signing authority. (required) :certfile path to the server certificate file. (required for Transport.Tls) :keyfile path to the signing key. (required for Transport.Tls) :customize_hostname_check it&#39;s very likely that you might get tls failures if you are relying on the OTP builtin hostname checks. This OTP ssl feature lets you override it for something custom. See :ssl.client_option/0 :reply_ttl the maximum amount of time that client should wait for call replies. Units in ms, defaults to 5000. forward_callers: true causes the client to adopt the universe of the caller. see Multiverses for details see GenServer.start_link/3 for a description of further options."},{"ref":"Erps.Client.html#c:terminate/2","title":"Erps.Client.terminate/2","type":"callback","doc":"Invoked when the client is about to exit. This would usually occur due to handle_push/2 returning a {:stop, reason, new_state} tuple, but also if the TCP connection happens to go down."},{"ref":"Erps.Client.html#t:noreply_response/0","title":"Erps.Client.noreply_response/0","type":"type","doc":""},{"ref":"Erps.Client.html#t:reply_cache/0","title":"Erps.Client.reply_cache/0","type":"type","doc":""},{"ref":"Erps.Client.html#t:reply_ref/0","title":"Erps.Client.reply_ref/0","type":"type","doc":""},{"ref":"Erps.Daemon.html","title":"Erps.Daemon","type":"module","doc":"A Process which sits on a TCP port and listens to inbound connections, handing them off to Erps.Server modules when an connection arrives. The Daemon also manages putting those server processes into supervision trees, usually a DynamicSupervisor. Each server will be seeded with a common initial data and common options, though they may incur specialization once handed off to the Erps.Server module&#39;s start_link/2 and init/1 functions. Example: this invocation will initialize the Erps.Server MyServer with an empty map, supervised by the DynamicSupervisor named ServerSupervisor. See start_link/2 for more supervision options. Erps.Daemon.start_link(MyServer, %{}, port: &lt;port_number&gt;, server_supervisor: ServerSupervisor) Typically, you will also want to supervise Erps.Daemon itself, in which case you should use the following form, passing it into a static Supervisor: children = [{Erps.Daemon, {MyServer, %{}, port: &lt;port_number&gt;, server_supervisor: ServerSupervisor}}] You may override the standard child_spec parameters [:id, :restart, :shutdown] in the options list of the tuple."},{"ref":"Erps.Daemon.html#child_spec/1","title":"Erps.Daemon.child_spec/1","type":"function","doc":""},{"ref":"Erps.Daemon.html#handle_call/3","title":"Erps.Daemon.handle_call/3","type":"function","doc":"Invoked to handle synchronous call/3 messages. call/3 will block until a reply is received (unless the call times out or nodes are disconnected). request is the request message sent by a call/3, from is a 2-tuple containing the caller&#39;s PID and a term that uniquely identifies the call, and state is the current state of the GenServer. Returning {:reply, reply, new_state} sends the response reply to the caller and continues the loop with new state new_state. Returning {:reply, reply, new_state, timeout} is similar to {:reply, reply, new_state} except that it also sets a timeout. See the &quot;Timeouts&quot; section in the module documentation for more information. Returning {:reply, reply, new_state, :hibernate} is similar to {:reply, reply, new_state} except the process is hibernated and will continue the loop once a message is in its message queue. If a message is already in the message queue this will be immediately. Hibernating a GenServer causes garbage collection and leaves a continuous heap that minimises the memory used by the process. Returning {:reply, reply, new_state, {:continue, continue}} is similar to {:reply, reply, new_state} except c:handle_continue/2 will be invoked immediately after with the value continue as first argument. Hibernating should not be used aggressively as too much time could be spent garbage collecting. Normally it should only be used when a message is not expected soon and minimising the memory of the process is shown to be beneficial. Returning {:noreply, new_state} does not send a response to the caller and continues the loop with new state new_state. The response must be sent with reply/2. There are three main use cases for not replying using the return value: To reply before returning from the callback because the response is known before calling a slow function. To reply after returning from the callback because the response is not yet available. To reply from another process, such as a task. When replying from another process the GenServer should exit if the other process exits without replying as the caller will be blocking awaiting a reply. Returning {:noreply, new_state, timeout | :hibernate | {:continue, continue}} is similar to {:noreply, new_state} except a timeout, hibernation or continue occurs as with a :reply tuple. Returning {:stop, reason, reply, new_state} stops the loop and c:terminate/2 is called with reason reason and state new_state. Then the reply is sent as the response to call and the process exits with reason reason. Returning {:stop, reason, new_state} is similar to {:stop, reason, reply, new_state} except a reply is not sent. This callback is optional. If one is not implemented, the server will fail if a call is performed against it. Callback implementation for GenServer.handle_call/3."},{"ref":"Erps.Daemon.html#handle_info/2","title":"Erps.Daemon.handle_info/2","type":"function","doc":"Invoked to handle all other messages. msg is the message and state is the current state of the GenServer. When a timeout occurs the message is :timeout. Return values are the same as c:handle_cast/2. This callback is optional. If one is not implemented, the received message will be logged. Callback implementation for GenServer.handle_info/2."},{"ref":"Erps.Daemon.html#port/1","title":"Erps.Daemon.port/1","type":"function","doc":"retrieve the TCP port that the erps daemon is bound to. Useful for tests - when we want to assign it a port of 0 so that it gets &quot;any free port&quot; of the system."},{"ref":"Erps.Daemon.html#start/3","title":"Erps.Daemon.start/3","type":"function","doc":"see start_link/2"},{"ref":"Erps.Daemon.html#start_link/3","title":"Erps.Daemon.start_link/3","type":"function","doc":"launches a Erps Daemon, linked to the calling process You can pass these general options which will propagate to the Erps.Servers. You may also want to specify the supervision tree using the :server_supervisor option as follows: server_supervisor: pid_or_name assumes the supervisor is DynamicSupervisor (or equivalent) and calls DynamicSupervisor.start_child/2 server_supervisor: {module, name} allows you to use a generic module for supervision and calls module.start_child/2 with the server module and the arity-2 parameters for the server_module.start_link/2 function. forward_callers: true causes the daemon and its spawned servers to adopt the universe of the caller. see Multiverses for details other options you may want to override: :port, sets the TCP port the daemon will listen on. Defaults to 0, which means that a random port will be selected, and must be retrieved using port/1 :transport, useful for mocking with TCP instead of TLS in tests :tls_opts, useful for specifiying TLS parameters shared by all server processes"},{"ref":"Erps.Daemon.html#t:socket/0","title":"Erps.Daemon.socket/0","type":"type","doc":""},{"ref":"Erps.Server.html","title":"Erps.Server","type":"behaviour","doc":"Create an Erps server GenServer. An Erps server is just a GenServer that has its call/2 and cast/2 callbacks connected to the external network over a transport portocol. Basic Operation Presuming you have set up TLS credentials, you can instantiate a server in basically the same way that you would instantiate a GenServer: defmodule ErpsServer do use Erps.Server @port &lt;...&gt; def start_link(data, opts) do Erps.Server.start_link(__MODULE__, data, opts) end @impl true def init(init_state), do: {:ok, init_state} def handle_call(:some_remote_call, state) do {:reply, :some_remote_response, state} end end NB you must implement a start_link/2 in order to be properly launched by Erps.Daemon. Now you may either access these values as a normal, local GenServer, or access them via Erps.Client. The Erps.Server module will not know the difference between a local or a remote call. {:ok, server} = ErpsServer.start_link() GenServer.call(server, :some_remote_call) #==&gt; :some_remote_response {:ok, client} = ErpsClient.start_link() GenServer.call(client, :some_remote_call) #==&gt; :some_remote_response Module Options :identifier a binary identifier for your Erps API endpoint. Maximum 12 bytes, suggested to be human-readable. :versions client semvers which are accepted. See the &quot;requirements&quot; section in Version. :safe (see :erlang.binary_to_term/2), for decoding terms. If set to false, then allows undefined atoms and lambdas to be passed via the protocol. This should be used with extreme caution, as disabling safe mode can be an attack vector. (defaults to true) :transport - set the transport module, which must implement Transport behaviour. If you set it to false, the Erps server will use Erps.Transport.None and act similarly to a GenServer (with some overhead). Example defmodule MyServer do use Erps.Server, versions: &quot;~&gt; 0.2.4&quot;, identifier: &quot;my_api&quot;, safe: false def start_link(iv) do Erps.Server.start_link(__MODULE__, init, port: , tls_opts: [...]) end def init(iv), do: {:ok, iv} end Magical features The following functions are hoisted to your server module so that you can call them in your code with clarity and less boilerplate: push/2 disconnect/1"},{"ref":"Erps.Server.html#child_spec/1","title":"Erps.Server.child_spec/1","type":"function","doc":""},{"ref":"Erps.Server.html#disconnect/1","title":"Erps.Server.disconnect/1","type":"function","doc":"disconnects the server, causing it to close."},{"ref":"Erps.Server.html#c:handle_call/3","title":"Erps.Server.handle_call/3","type":"callback","doc":"similar to GenServer.handle_call/3, but handles content from both local and remote clients. The from term might contain a opaque term which represents a return address for a remote client, but you may use this term as expected in reply/2. Return codes {:reply, reply, new_state} replies, then updates the state of the Server. {:reply, reply, new_state, timeout} replies, then causes a :timeout message to be sent to handle_info/2 if no other message comes by {:reply, reply, new_state, :hibernate} replies, then causes a hibernation event (see :erlang.hibernate/3) {:reply, reply, new_state, {:continue, term}} replies, then causes a continuation to be triggered after the message is handled, it will be sent to c:handle_continue/3 and all return codes supported by GenServer.handle_cast/2"},{"ref":"Erps.Server.html#c:handle_cast/2","title":"Erps.Server.handle_cast/2","type":"callback","doc":"similar to GenServer.handle_cast/2, but handles content from both local and remote clients (if the content has been successfully filtered). Return codes {:noreply, new_state} continues the loop with new state new_state {:noreply, new_state, timeout} causes a :timeout message to be sent to handle_info/2 if no other message comes by {:noreply, new_state, :hibernate}, causes a hibernation event (see :erlang.hibernate/3) {:noreply, new_state, {:continue, term}} causes a continuation to be triggered after the message is handled, it will be sent to c:handle_continue/3 {:stop, reason, new_state} terminates the loop, passing new_state to terminate/2, if it&#39;s implemented."},{"ref":"Erps.Server.html#c:handle_continue/2","title":"Erps.Server.handle_continue/2","type":"callback","doc":"Invoked when an internal callback requests a continuation, using {:noreply, state, {:continue, continuation}}, or from init/1 using {:ok, state, {:continue, continuation}} see: GenServer.handle_continue/2. Return codes see return codes for handle_cast/2"},{"ref":"Erps.Server.html#c:handle_info/2","title":"Erps.Server.handle_info/2","type":"callback","doc":"Invoked to handle general messages sent to the client process. Most useful if the client needs to be attentive to system messages, such as nodedown or monitored processes, but also useful for internal timeouts. see: GenServer.handle_info/2. Return codes see return codes for handle_cast/2"},{"ref":"Erps.Server.html#c:init/1","title":"Erps.Server.init/1","type":"callback","doc":"Invoked to set up the process. Like GenServer.init/1, this function is called from inside the process immediately after start_link/3 or start/3. Return codes {:ok, state} a succesful startup of your intialization logic and sets the internal state of your server to state. {:ok, state, timeout} the above, plus a :timeout atom will be sent to handle_info/2 if no other messages come by. {:ok, state, :hibernate} successful startup, followed by a hibernation event (see :erlang.hibernate/3) {:ok, state, {:continue, term}} successful startup, and causes a continuation to be triggered after the message is handled, sent to c:handle_continue/3 :ignore - Drop the gen_server creation request, because for some reason it shouldn&#39;t have started. {:stop, reason} - a failure in creating the gen_server. Results in {:error, reason} being propagated as the result of the start_link"},{"ref":"Erps.Server.html#push/2","title":"Erps.Server.push/2","type":"function","doc":"pushes a message to all connected clients. Causes client Erps.Client.handle_push/2 callbacks to be triggered."},{"ref":"Erps.Server.html#reply/2","title":"Erps.Server.reply/2","type":"function","doc":"sends a reply to either a local process or a remotely connected client. The Erps Server holds connection information, so you must supply the Erps Server pid; though you may use the arity-2 form if you call from within the action loop of the Erps Server itself, in which case it acts like GenServer.reply/2 naturally takes the from value passed in to the second parameter of handle_call/3."},{"ref":"Erps.Server.html#start/3","title":"Erps.Server.start/3","type":"function","doc":"starts a server GenServer, not linked to the caller. Most useful for tests. see start_link/3 for a description of avaliable options."},{"ref":"Erps.Server.html#start_link/3","title":"Erps.Server.start_link/3","type":"function","doc":"starts a server GenServer, linked to the caller. options :transport transport module (see Transport.Api) :tls_opts options for TLS authorization and encryption. Should include: :cacertfile path to the certificate of your signing authority. :certfile path to the server certificate file. :keyfile path to the signing key. :customize_hostname_check see below. customize_hostname_check Erlang/OTP doesn&#39;t provide a simple mechanism for a server in a two way ssl connection to verify the identity of the client (which you may want in certain situations where you own both ends of the connection but not the intermediate transport layer). See Transport.Tls.handshake/2 and :public_key.pkix_verify_hostname/3 for more details. see GenServer.start_link/3 for a description of further options."},{"ref":"Erps.Server.html#c:terminate/2","title":"Erps.Server.terminate/2","type":"callback","doc":"see: GenServer.terminate/2."},{"ref":"Erps.Server.html#t:from/0","title":"Erps.Server.from/0","type":"opaque","doc":"a from term that is either a local from, compatible with GenServer.from/0 or an opaque term that represents a connected remote client. This opaque term is completely compatible with the t:GenServer.from type. In other words, you may pass this term to Genserver.reply/2 from any process and it will correctly route it to the caller whether it be local or remote."},{"ref":"Erps.Server.html#t:server/0","title":"Erps.Server.server/0","type":"type","doc":""},{"ref":"Erps.Transport.None.html","title":"Erps.Transport.None","type":"module","doc":"implements no transport, which turns the Erps server into a basic GenServer with some overhead."},{"ref":"Erps.Transport.None.html#recv/3","title":"Erps.Transport.None.recv/3","type":"function","doc":"Callback implementation for Transport.recv/3, via :ssl.recv/3."}]